import re

from dagster_graphql.test.utils import execute_dagster_graphql, infer_pipeline_selector

from .graphql_context_test_suite import NonLaunchableGraphQLContextTestMatrix

SCHEMA_OR_ERROR_SUBSET_QUERY = """
query EnvironmentQuery($selector: PipelineSelector!){
    runConfigSchemaOrError(selector: $selector) {
        __typename
        ... on RunConfigSchema {
            allConfigTypes {
                __typename
                key
                ... on CompositeConfigType {
                    __typename
                    fields {
                        __typename
                        name
                        configType {
                            key
                            __typename
                        }
                    }
                }
            }
        }
        ... on InvalidSubsetError {
            message
        }
        ... on PythonError {
            message
            stack
        }
    }
}
"""


def field_names_of(type_dict, typename):
    return {field_data["name"] for field_data in type_dict[typename]["fields"]}


def types_dict_of_result(subset_result, top_key):
    return {
        type_data["name"]: type_data for type_data in subset_result.data[top_key]["configTypes"]
    }


class TestSolidSelections(NonLaunchableGraphQLContextTestMatrix):
    def test_csv_hello_world_pipeline_or_error_subset_wrong_solid_name(self, graphql_context):
        selector = infer_pipeline_selector(graphql_context, "csv_hello_world", ["nope"])
        result = execute_dagster_graphql(
            graphql_context, SCHEMA_OR_ERROR_SUBSET_QUERY, {"selector": selector}
        )

        assert not result.errors
        assert result.data
        assert result.data["runConfigSchemaOrError"]["__typename"] == "InvalidSubsetError"
        assert "No qualified solids to execute" in result.data["runConfigSchemaOrError"]["message"]

    def test_pipeline_with_invalid_definition_error(self, graphql_context):
        selector = infer_pipeline_selector(
            graphql_context, "pipeline_with_invalid_definition_error", ["fail_subset"]
        )
        result = execute_dagster_graphql(
            graphql_context, SCHEMA_OR_ERROR_SUBSET_QUERY, {"selector": selector}
        )
        assert not result.errors
        assert result.data
        assert result.data["runConfigSchemaOrError"]["__typename"] == "InvalidSubsetError"

        assert re.match(
            (
                r".*DagsterInvalidSubsetError[\s\S]*"
                r"add a dagster_type_loader for the type 'InputTypeWithoutHydration'"
            ),
            result.data["runConfigSchemaOrError"]["message"],
        )



from collections import defaultdict
from typing import Dict, List

from graphql.execution.base import ResolveInfo

from dagster import PipelineDefinition, PipelineRunStatus, check
from dagster.config.validate import validate_config
from dagster.core.definitions import create_run_config_schema
from dagster.core.errors import DagsterRunNotFoundError
from dagster.core.execution.stats import StepEventStatus
from dagster.core.host_representation import PipelineSelector
from dagster.core.storage.pipeline_run import PipelineRun, RunsFilter
from dagster.core.storage.tags import TagType, get_tag_type
from dagster.utils import utc_datetime_from_timestamp

from .external import ensure_valid_config, get_external_pipeline_or_raise
from .utils import UserFacingGraphQLError, capture_error


def is_config_valid(pipeline_def, run_config, mode):
    check.str_param(mode, "mode")
    check.inst_param(pipeline_def, "pipeline_def", PipelineDefinition)

    run_config_schema = create_run_config_schema(pipeline_def, mode)
    validated_config = validate_config(run_config_schema.config_type, run_config)
    return validated_config.success


def get_validated_config(pipeline_def, run_config, mode):
    from ..schema.pipelines.config import GrapheneRunConfigValidationInvalid

    check.str_param(mode, "mode")
    check.inst_param(pipeline_def, "pipeline_def", PipelineDefinition)

    run_config_schema = create_run_config_schema(pipeline_def, mode)

    validated_config = validate_config(run_config_schema.config_type, run_config)

    if not validated_config.success:
        raise UserFacingGraphQLError(
            GrapheneRunConfigValidationInvalid.for_validation_errors(
                pipeline_def.get_external_pipeline(), validated_config.errors
            )
        )

    return validated_config


def get_run_by_id(graphene_info, run_id):
    from ..schema.errors import GrapheneRunNotFoundError
    from ..schema.pipelines.pipeline import GrapheneRun

    instance = graphene_info.context.instance
    records = instance.get_run_records(RunsFilter(run_ids=[run_id]))
    if not records:
        return GrapheneRunNotFoundError(run_id)
    else:
        return GrapheneRun(records[0])


def get_run_tags(graphene_info):
    from ..schema.tags import GraphenePipelineTagAndValues

    instance = graphene_info.context.instance
    return [
        GraphenePipelineTagAndValues(key=key, values=values)
        for key, values in instance.get_run_tags()
        if get_tag_type(key) != TagType.HIDDEN
    ]


@capture_error
def get_run_group(graphene_info, run_id):
    from ..schema.errors import GrapheneRunGroupNotFoundError
    from ..schema.pipelines.pipeline import GrapheneRun
    from ..schema.runs import GrapheneRunGroup

    instance = graphene_info.context.instance
    try:
        result = instance.get_run_group(run_id)
    except DagsterRunNotFoundError:
        return GrapheneRunGroupNotFoundError(run_id)
    root_run_id, run_group = result
    run_group_run_ids = [run.run_id for run in run_group]
    records_by_id = {
        record.pipeline_run.run_id: record
        for record in instance.get_run_records(RunsFilter(run_ids=run_group_run_ids))
    }
    return GrapheneRunGroup(
        root_run_id=root_run_id,
        runs=[GrapheneRun(records_by_id.get(run_id)) for run_id in run_group_run_ids],
    )


def get_runs(graphene_info, filters, cursor=None, limit=None):
    from ..schema.pipelines.pipeline import GrapheneRun

    check.opt_inst_param(filters, "filters", RunsFilter)
    check.opt_str_param(cursor, "cursor")
    check.opt_int_param(limit, "limit")

    instance = graphene_info.context.instance

    return [
        GrapheneRun(record)
        for record in instance.get_run_records(filters=filters, cursor=cursor, limit=limit)
    ]


PENDING_STATUSES = [
    PipelineRunStatus.STARTING,
    PipelineRunStatus.MANAGED,
    PipelineRunStatus.NOT_STARTED,
    PipelineRunStatus.QUEUED,
    PipelineRunStatus.STARTED,
    PipelineRunStatus.CANCELING,
]
IN_PROGRESS_STATUSES = [
    PipelineRunStatus.STARTED,
    PipelineRunStatus.CANCELING,
]


def get_in_progress_runs_by_step(graphene_info, job_names, step_keys):
    from ..schema.pipelines.pipeline import GrapheneInProgressRunsByStep, GrapheneRun

    instance = graphene_info.context.instance

    in_progress_records = []
    for job_name in job_names:
        in_progress_records.extend(
            instance.get_run_records(RunsFilter(pipeline_name=job_name, statuses=PENDING_STATUSES))
        )

    in_progress_runs_by_step = defaultdict(list)
    unstarted_runs_by_step = defaultdict(list)

    for record in in_progress_records:
        run = record.pipeline_run

        asset_names = graphene_info.context.instance.get_execution_plan_snapshot(
            run.execution_plan_snapshot_id
        ).step_keys_to_execute

        if run.status in IN_PROGRESS_STATUSES:
            step_stats = graphene_info.context.instance.get_run_step_stats(run.run_id, step_keys)
            for step_stat in step_stats:
                if step_stat.status == StepEventStatus.IN_PROGRESS:
                    in_progress_runs_by_step[step_stat.step_key].append(GrapheneRun(record))

            for step_key in asset_names:
                # step_stats only contains stats for steps that are in progress or complete
                is_unstarted = (
                    len([step_stat for step_stat in step_stats if step_stat.step_key == step_key])
                    == 0
                )
                if is_unstarted:
                    unstarted_runs_by_step[step_key].append(GrapheneRun(record))
        else:
            # the run never began execution, all steps are unstarted
            for step_key in asset_names:
                unstarted_runs_by_step[step_key].append(GrapheneRun(record))

    all_step_keys = in_progress_runs_by_step.keys() | unstarted_runs_by_step.keys()
    return [
        GrapheneInProgressRunsByStep(
            key,
            unstarted_runs_by_step.get(key, []),
            in_progress_runs_by_step.get(key, []),
        )
        for key in all_step_keys
    ]


def get_asset_run_stats_by_step(graphene_info, job_names, asset_nodes):
    # This is a utility method that gets the latest run that selected an asset,
    # by searching within the last 5 runs of each job the asset belongs in.
    # If none of the most recent runs selected the asset, we return back a GrapheneJobRunsCount
    # object that contains the total number of job runs that have occurred since the latest
    # asset materialization (or the total number of job runs if the asset has never been
    # materialized).

    latest_run_by_step = get_latest_asset_run_by_step_key(graphene_info, job_names, asset_nodes)
    assets_to_fetch_run_count = [
        asset_node
        for asset_node in asset_nodes
        if latest_run_by_step.get(asset_node.op_name) == None
    ]
    jobs_runs_count = get_asset_runs_count_by_step(graphene_info, assets_to_fetch_run_count)

    return [
        latest_run_by_step.get(asset_node.op_name, jobs_runs_count.get(asset_node.op_name))
        for asset_node in asset_nodes
    ]


def get_latest_asset_run_by_step_key(graphene_info, job_names, asset_nodes):
    from ..schema.pipelines.pipeline import GrapheneLatestRun, GrapheneRun

    # This method returns the latest run that has occurred for a given step.
    # Because it is expensive to deserialize PipelineRun objects, we limit this
    # query to retrieving the last 5 runs per job. If no runs have occurred, we return
    # a GrapheneLatestRun object with no run. If none of the latest runs contain the
    # step key, we return None.

    instance = graphene_info.context.instance

    latest_run_by_step: Dict[str, PipelineRun] = {}

    run_records_by_job = {
        job_name: instance.get_run_records(RunsFilter(pipeline_name=job_name), limit=5)
        for job_name in job_names
    }

    get_step_keys = lambda record: graphene_info.context.instance.get_execution_plan_snapshot(
        record.pipeline_run.execution_plan_snapshot_id
    ).step_keys_to_execute

    # step_keys list is in same order as records list in get_step_keys
    record_step_keys_by_job = {
        job_name: [get_step_keys(record) for record in run_records]
        for job_name, run_records in run_records_by_job.items()
    }

    for asset_node in asset_nodes:
        asset_job_names = asset_node.job_names
        step_key = asset_node.op_name

        total_num_records = sum(
            [len(run_records_by_job.get(job_name, [])) for job_name in asset_job_names]
        )
        if total_num_records == 0:
            latest_run_by_step[step_key] = GrapheneLatestRun(step_key, None)

        latest_run = None
        for job_name in asset_job_names:
            record_step_keys = record_step_keys_by_job.get(job_name, [])
            for i in range(len(record_step_keys)):
                if step_key in record_step_keys[i]:
                    record = run_records_by_job.get(job_name)[i]
                    if latest_run == None or record.create_timestamp > latest_run.create_timestamp:
                        latest_run = record

        if latest_run:
            latest_run_by_step[step_key] = GrapheneLatestRun(step_key, GrapheneRun(latest_run))

    return latest_run_by_step


def get_asset_runs_count_by_step(graphene_info, asset_nodes):
    from ..schema.pipelines.pipeline import GrapheneJobRunsCount

    instance = graphene_info.context.instance

    jobs_runs_count: Dict[str, GrapheneJobRunsCount] = {}

    if len(asset_nodes) == 0:
        return jobs_runs_count

    step_key_to_job_names: Dict[str, List[str]] = {
        asset_node.op_name: asset_node.job_names for asset_node in asset_nodes
    }
    materializations = instance.get_latest_materialization_events(
        [asset_node.asset_key for asset_node in asset_nodes]
    )
    for asset_node in asset_nodes:
        event = materializations.get(asset_node.asset_key)
        step_key = asset_node.op_name
        job_names = step_key_to_job_names[step_key]
        runs_count = sum(
            [
                instance.get_runs_count(
                    RunsFilter(
                        pipeline_name=job_name,
                        updated_after=utc_datetime_from_timestamp(event.timestamp)
                        if event
                        else None,
                    )
                )
                for job_name in job_names
            ]
        )

        # If a materialization has occurred, we subtract one so that the runs count
        # does not include the run that generated the materialization.
        if event:
            runs_count -= 1

        jobs_runs_count[step_key] = GrapheneJobRunsCount(
            step_key, job_names, runs_count, True if event else False
        )
    return jobs_runs_count


def get_runs_count(graphene_info, filters):
    return graphene_info.context.instance.get_runs_count(filters)


def get_run_groups(graphene_info, filters=None, cursor=None, limit=None):
    from ..schema.pipelines.pipeline import GrapheneRun
    from ..schema.runs import GrapheneRunGroup

    check.opt_inst_param(filters, "filters", RunsFilter)
    check.opt_str_param(cursor, "cursor")
    check.opt_int_param(limit, "limit")

    instance = graphene_info.context.instance
    run_groups = instance.get_run_groups(filters=filters, cursor=cursor, limit=limit)
    run_ids = {run.run_id for run_group in run_groups.values() for run in run_group.get("runs", [])}
    records_by_ids = {
        record.pipeline_run.run_id: record
        for record in instance.get_run_records(RunsFilter(run_ids=list(run_ids)))
    }

    for root_run_id in run_groups:
        run_groups[root_run_id]["runs"] = [
            GrapheneRun(records_by_ids.get(run.run_id)) for run in run_groups[root_run_id]["runs"]
        ]

    return [
        GrapheneRunGroup(root_run_id=root_run_id, runs=run_group["runs"])
        for root_run_id, run_group in run_groups.items()
    ]


@capture_error
def validate_pipeline_config(graphene_info, selector, run_config, mode):
    from ..schema.pipelines.config import GraphenePipelineConfigValidationValid

    check.inst_param(graphene_info, "graphene_info", ResolveInfo)
    check.inst_param(selector, "selector", PipelineSelector)
    check.opt_str_param(mode, "mode")

    external_pipeline = get_external_pipeline_or_raise(graphene_info, selector)
    ensure_valid_config(external_pipeline, mode, run_config)
    return GraphenePipelineConfigValidationValid(pipeline_name=external_pipeline.name)


@capture_error
def get_execution_plan(graphene_info, selector, run_config, mode):
    from ..schema.execution import GrapheneExecutionPlan

    check.inst_param(graphene_info, "graphene_info", ResolveInfo)
    check.inst_param(selector, "selector", PipelineSelector)
    check.opt_str_param(mode, "mode")

    external_pipeline = get_external_pipeline_or_raise(graphene_info, selector)
    ensure_valid_config(external_pipeline, mode, run_config)
    return GrapheneExecutionPlan(
        graphene_info.context.get_external_execution_plan(
            external_pipeline=external_pipeline,
            mode=mode,
            run_config=run_config,
            step_keys_to_execute=None,
            known_state=None,
        )
    )


@capture_error
def get_stats(graphene_info, run_id):
    from ..schema.pipelines.pipeline_run_stats import GrapheneRunStatsSnapshot

    stats = graphene_info.context.instance.get_run_stats(run_id)
    stats.id = "stats-{run_id}"
    return GrapheneRunStatsSnapshot(stats)


def get_step_stats(graphene_info, run_id, step_keys=None):
    from ..schema.logs.events import GrapheneRunStepStats

    step_stats = graphene_info.context.instance.get_run_step_stats(run_id, step_keys)
    return [GrapheneRunStepStats(stats) for stats in step_stats]



import sys
from collections import namedtuple
from typing import cast

from graphql.execution.base import ResolveInfo

from dagster import check
from dagster.core.host_representation import GraphSelector, PipelineSelector
from dagster.core.workspace.context import BaseWorkspaceRequestContext
from dagster.utils.error import serializable_error_info_from_exc_info


def check_permission(permission):
    def decorator(fn):
        def _fn(self, graphene_info, *args, **kwargs):  # pylint: disable=unused-argument
            assert_permission(graphene_info, permission)

            return fn(self, graphene_info, *args, **kwargs)

        return _fn

    return decorator


def assert_permission(graphene_info: ResolveInfo, permission: str) -> None:
    from dagster_graphql.schema.errors import GrapheneUnauthorizedError

    context = cast(BaseWorkspaceRequestContext, graphene_info.context)
    if not context.has_permission(permission):
        raise UserFacingGraphQLError(GrapheneUnauthorizedError())


class ErrorCapture:
    @staticmethod
    def default_on_exception(exc_info):
        from dagster_graphql.schema.errors import GraphenePythonError

        # Transform exception in to PythonError to present to user
        return GraphenePythonError(serializable_error_info_from_exc_info(exc_info))

    on_exception = default_on_exception


def capture_error(fn):
    def _fn(*args, **kwargs):

        try:
            return fn(*args, **kwargs)
        except UserFacingGraphQLError as de_exception:
            return de_exception.error
        except Exception:
            return ErrorCapture.on_exception(sys.exc_info())

    return _fn


class UserFacingGraphQLError(Exception):
    def __init__(self, error):
        self.error = error
        message = "[{cls}] {message}".format(
            cls=error.__class__.__name__,
            message=error.message if hasattr(error, "message") else None,
        )
        super(UserFacingGraphQLError, self).__init__(message)


def pipeline_selector_from_graphql(data):
    return PipelineSelector(
        location_name=data["repositoryLocationName"],
        repository_name=data["repositoryName"],
        pipeline_name=data.get("pipelineName") or data.get("jobName"),
        solid_selection=data.get("solidSelection"),
    )


def graph_selector_from_graphql(data):
    return GraphSelector(
        location_name=data["repositoryLocationName"],
        repository_name=data["repositoryName"],
        graph_name=data["graphName"],
    )


class ExecutionParams(
    namedtuple(
        "_ExecutionParams",
        "selector run_config mode execution_metadata step_keys",
    )
):
    def __new__(cls, selector, run_config, mode, execution_metadata, step_keys):
        check.dict_param(run_config, "run_config", key_type=str)
        check.opt_list_param(step_keys, "step_keys", of_type=str)

        return super(ExecutionParams, cls).__new__(
            cls,
            selector=check.inst_param(selector, "selector", PipelineSelector),
            run_config=run_config,
            mode=check.opt_str_param(mode, "mode"),
            execution_metadata=check.inst_param(
                execution_metadata, "execution_metadata", ExecutionMetadata
            ),
            step_keys=step_keys,
        )

    def to_graphql_input(self):
        return {
            "selector": self.selector.to_graphql_input(),
            "runConfigData": self.run_config,
            "mode": self.mode,
            "executionMetadata": self.execution_metadata.to_graphql_input(),
            "stepKeys": self.step_keys,
        }


class ExecutionMetadata(namedtuple("_ExecutionMetadata", "run_id tags root_run_id parent_run_id")):
    def __new__(cls, run_id, tags, root_run_id=None, parent_run_id=None):
        return super(ExecutionMetadata, cls).__new__(
            cls,
            check.opt_str_param(run_id, "run_id"),
            check.dict_param(tags, "tags", key_type=str, value_type=str),
            check.opt_str_param(root_run_id, "root_run_id"),
            check.opt_str_param(parent_run_id, "parent_run_id"),
        )

    def to_graphql_input(self):
        return {
            "runId": self.run_id,
            "tags": [{"key": k, "value": v} for k, v in self.tags.items()],
            "rootRunId": self.root_run_id,
            "parentRunId": self.parent_run_id,
        }



import time

import pytest
from dagster_graphql import DagsterGraphQLClientError
from dagster_graphql.client.query import LAUNCH_PIPELINE_EXECUTION_MUTATION
from dagster_graphql.test.utils import execute_dagster_graphql, infer_pipeline_selector

from dagster.core.storage.pipeline_run import PipelineRunStatus

from ..graphql.graphql_context_test_suite import ExecutingGraphQLContextTestMatrix
from ..graphql.setup import csv_hello_world_solids_config
from .conftest import MockClient, python_client_test_suite


@python_client_test_suite
def test_get_run_status_success(mock_client: MockClient):
    expected_result = PipelineRunStatus.SUCCESS
    response = {"pipelineRunOrError": {"__typename": "PipelineRun", "status": expected_result}}
    mock_client.mock_gql_client.execute.return_value = response

    actual_result = mock_client.python_client.get_run_status("foo")
    assert actual_result == expected_result


@python_client_test_suite
def test_get_run_status_fails_with_python_error(mock_client: MockClient):
    error_type, error_msg = "PythonError", "something exploded"
    response = {"pipelineRunOrError": {"__typename": error_type, "message": error_msg}}
    mock_client.mock_gql_client.execute.return_value = response

    with pytest.raises(DagsterGraphQLClientError) as exc_info:
        mock_client.python_client.get_run_status("foo")

    assert exc_info.value.args == (error_type, error_msg)


@python_client_test_suite
def test_get_run_status_fails_with_pipeline_run_not_found_error(mock_client: MockClient):
    error_type, error_msg = "RunNotFoundError", "The specified pipeline run does not exist"
    response = {"pipelineRunOrError": {"__typename": error_type, "message": error_msg}}
    mock_client.mock_gql_client.execute.return_value = response

    with pytest.raises(DagsterGraphQLClientError) as exc_info:
        mock_client.python_client.get_run_status("foo")

    assert exc_info.value.args == (error_type, error_msg)


@python_client_test_suite
def test_get_run_status_fails_with_query_error(mock_client: MockClient):
    mock_client.mock_gql_client.execute.side_effect = Exception("foo")

    with pytest.raises(DagsterGraphQLClientError) as _:
        mock_client.python_client.get_run_status("foo")


class TestGetRunStatusWithClient(ExecutingGraphQLContextTestMatrix):
    def test_get_run_status(self, graphql_context, graphql_client):
        selector = infer_pipeline_selector(graphql_context, "csv_hello_world")
        result = execute_dagster_graphql(
            graphql_context,
            LAUNCH_PIPELINE_EXECUTION_MUTATION,
            variables={
                "executionParams": {
                    "selector": selector,
                    "runConfigData": csv_hello_world_solids_config(),
                    "mode": "default",
                }
            },
        )

        assert not result.errors
        assert result.data

        run_id = result.data["launchPipelineExecution"]["run"]["runId"]

        start_time = time.time()

        while True:
            if time.time() - start_time > 30:
                raise Exception("Timed out waiting for get_run_status to return SUCCESS")

            status = graphql_client.get_run_status(run_id)

            if status == PipelineRunStatus.SUCCESS:
                break

            time.sleep(3)



import graphene

from ..errors import GraphenePipelineNotFoundError, GraphenePythonError
from .config import GraphenePipelineConfigValidationValid, GrapheneRunConfigValidationInvalid
from .pipeline_errors import GrapheneInvalidSubsetError


class GraphenePipelineConfigValidationResult(graphene.Union):
    class Meta:
        types = (
            GrapheneInvalidSubsetError,
            GraphenePipelineConfigValidationValid,
            GrapheneRunConfigValidationInvalid,
            GraphenePipelineNotFoundError,
            GraphenePythonError,
        )
        name = "PipelineConfigValidationResult"



import graphene

from dagster import check
from dagster.core.host_representation import RepresentedPipeline

from ..implementation.run_config_schema import resolve_is_run_config_valid
from .config_types import GrapheneConfigType, to_config_type
from .errors import GrapheneModeNotFoundError, GraphenePipelineNotFoundError, GraphenePythonError
from .pipelines.config_result import GraphenePipelineConfigValidationResult
from .runs import GrapheneRunConfigData, parse_run_config_input
from .util import non_null_list


class GrapheneRunConfigSchema(graphene.ObjectType):
    rootConfigType = graphene.Field(
        graphene.NonNull(GrapheneConfigType),
        description="""Fetch the root environment type. Concretely this is the type that
        is in scope at the root of configuration document for a particular execution selection.
        It is the type that is in scope initially with a blank config editor.""",
    )
    allConfigTypes = graphene.Field(
        non_null_list(GrapheneConfigType),
        description="""Fetch all the named config types that are in the schema. Useful
        for things like a type browser UI, or for fetching all the types are in the
        scope of a document so that the index can be built for the autocompleting editor.
    """,
    )

    isRunConfigValid = graphene.Field(
        graphene.NonNull(GraphenePipelineConfigValidationResult),
        args={"runConfigData": graphene.Argument(GrapheneRunConfigData)},
        description="""Parse a particular run config result. The return value
        either indicates that the validation succeeded by returning
        `PipelineConfigValidationValid` or that there are configuration errors
        by returning `RunConfigValidationInvalid' which containers a list errors
        so that can be rendered for the user""",
    )

    class Meta:
        description = """The run config schema represents the all the config type
        information given a certain execution selection and mode of execution of that
        selection. All config interactions (e.g. checking config validity, fetching
        all config types, fetching in a particular config type) should be done
        through this type """
        name = "RunConfigSchema"

    def __init__(self, represented_pipeline, mode):
        super().__init__()
        self._represented_pipeline = check.inst_param(
            represented_pipeline, "represented_pipeline", RepresentedPipeline
        )
        self._mode = check.str_param(mode, "mode")

    def resolve_allConfigTypes(self, _graphene_info):
        return sorted(
            list(
                map(
                    lambda key: to_config_type(
                        self._represented_pipeline.config_schema_snapshot, key
                    ),
                    self._represented_pipeline.config_schema_snapshot.all_config_keys,
                )
            ),
            key=lambda ct: ct.key,
        )

    def resolve_rootConfigType(self, _graphene_info):
        return to_config_type(
            self._represented_pipeline.config_schema_snapshot,
            self._represented_pipeline.get_mode_def_snap(self._mode).root_config_key,
        )

    def resolve_isRunConfigValid(self, graphene_info, **kwargs):
        return resolve_is_run_config_valid(
            graphene_info,
            self._represented_pipeline,
            self._mode,
            parse_run_config_input(kwargs.get("runConfigData", {})),
        )


class GrapheneRunConfigSchemaOrError(graphene.Union):
    class Meta:
        from .pipelines.pipeline_errors import GrapheneInvalidSubsetError

        types = (
            GrapheneRunConfigSchema,
            GraphenePipelineNotFoundError,
            GrapheneInvalidSubsetError,
            GrapheneModeNotFoundError,
            GraphenePythonError,
        )
        name = "RunConfigSchemaOrError"



import time

from dagster_graphql.test.utils import define_out_of_process_context, execute_dagster_graphql

from dagster import PresetDefinition, pipeline, repository, solid
from dagster.core.storage.pipeline_run import PipelineRunStatus
from dagster.core.test_utils import instance_for_test

RUNS_QUERY = """
query RunsQuery {
  pipelineRunsOrError {
    __typename
    ... on PipelineRuns {
      results {
        runId
        pipelineName
        status
        runConfigYaml
        stats {
          ... on PipelineRunStatsSnapshot {
            startTime
            endTime
            stepsFailed
          }
        }
      }
    }
  }
}
"""

PAGINATED_RUNS_QUERY = """
query PaginatedRunsQuery($cursor: String!, $limit: Int) {
  pipelineRunsOrError(
    cursor: $cursor
    limit: $limit
  ) {
    __typename
    ... on PipelineRuns {
      results {
        runId
        pipelineName
        status
        runConfigYaml
        stats {
          ... on PipelineRunStatsSnapshot {
            startTime
            endTime
            stepsFailed
          }
        }
      }
    }
  }
}
"""

FILTERED_RUNS_QUERY = """
query FilteredRunsQuery {
  pipelineRunsOrError(filter: { statuses: [FAILURE] }) {
    __typename
    ... on PipelineRuns {
      results {
        runId
        pipelineName
        status
        runConfigYaml
        stats {
          ... on PipelineRunStatsSnapshot {
            startTime
            endTime
            stepsFailed
          }
        }
      }
    }
  }
}
"""

REPOSITORIES_QUERY = """
query RepositoriesQuery {
  repositoriesOrError {
    ... on RepositoryConnection {
      nodes {
        name
        location {
          name
        }
      }
    }
  }
}
"""

PIPELINES_QUERY = """
query PipelinesQuery(
  $repositoryLocationName: String!
  $repositoryName: String!
) {
  repositoryOrError(
    repositorySelector: {
      repositoryLocationName: $repositoryLocationName
      repositoryName: $repositoryName
    }
  ) {
    ... on Repository {
      pipelines {
        name
      }
    }
  }
}
"""

LAUNCH_PIPELINE = """
mutation ExecutePipeline(
  $repositoryLocationName: String!
  $repositoryName: String!
  $pipelineName: String!
  $runConfigData: RunConfigData!
  $mode: String!
) {
  launchPipelineExecution(
    executionParams: {
      selector: {
        repositoryLocationName: $repositoryLocationName
        repositoryName: $repositoryName
        pipelineName: $pipelineName
      }
      runConfigData: $runConfigData
      mode: $mode
    }
  ) {
    __typename
    ... on LaunchPipelineRunSuccess {
      run {
        runId
      }
    }
    ... on PipelineConfigValidationInvalid {
      errors {
        message
        reason
      }
    }
    ... on PythonError {
      message
    }
  }
}
"""

LAUNCH_PIPELINE_PRESET = """
mutation ExecutePipeline(
  $repositoryLocationName: String!
  $repositoryName: String!
  $pipelineName: String!
  $presetName: String!
) {
  launchPipelineExecution(
    executionParams: {
      selector: {
        repositoryLocationName: $repositoryLocationName
        repositoryName: $repositoryName
        pipelineName: $pipelineName
      }
      preset: $presetName
    }
  ) {
    __typename
    ... on LaunchPipelineRunSuccess {
      run {
        runId
      }
    }
  }
}
"""


def get_repo():
    @solid
    def my_solid():
        pass

    @solid
    def loop():
        while True:
            time.sleep(0.1)

    @pipeline
    def infinite_loop_pipeline():
        loop()

    @pipeline(preset_defs=[PresetDefinition(name="my_preset", run_config={})])
    def foo_pipeline():
        my_solid()

    @repository
    def my_repo():
        return [infinite_loop_pipeline, foo_pipeline]

    return my_repo


def test_runs_query():
    with instance_for_test() as instance:
        repo = get_repo()
        run_id_1 = instance.create_run_for_pipeline(
            repo.get_pipeline("foo_pipeline"), status=PipelineRunStatus.STARTED
        ).run_id
        run_id_2 = instance.create_run_for_pipeline(
            repo.get_pipeline("foo_pipeline"), status=PipelineRunStatus.FAILURE
        ).run_id
        with define_out_of_process_context(__file__, "get_repo", instance) as context:
            result = execute_dagster_graphql(context, RUNS_QUERY)
            assert result.data
            run_ids = [run["runId"] for run in result.data["pipelineRunsOrError"]["results"]]
            assert len(run_ids) == 2
            assert run_ids[0] == run_id_2
            assert run_ids[1] == run_id_1


def test_paginated_runs_query():
    with instance_for_test() as instance:
        repo = get_repo()
        _ = instance.create_run_for_pipeline(
            repo.get_pipeline("foo_pipeline"), status=PipelineRunStatus.STARTED
        ).run_id
        run_id_2 = instance.create_run_for_pipeline(
            repo.get_pipeline("foo_pipeline"), status=PipelineRunStatus.FAILURE
        ).run_id
        run_id_3 = instance.create_run_for_pipeline(
            repo.get_pipeline("foo_pipeline"), status=PipelineRunStatus.SUCCESS
        ).run_id
        with define_out_of_process_context(__file__, "get_repo", instance) as context:
            result = execute_dagster_graphql(
                context,
                PAGINATED_RUNS_QUERY,
                variables={"cursor": run_id_3, "limit": 1},
            )
            assert result.data
            run_ids = [run["runId"] for run in result.data["pipelineRunsOrError"]["results"]]
            assert len(run_ids) == 1
            assert run_ids[0] == run_id_2


def test_filtered_runs_query():
    with instance_for_test() as instance:
        repo = get_repo()
        _ = instance.create_run_for_pipeline(
            repo.get_pipeline("foo_pipeline"), status=PipelineRunStatus.STARTED
        ).run_id
        run_id_2 = instance.create_run_for_pipeline(
            repo.get_pipeline("foo_pipeline"), status=PipelineRunStatus.FAILURE
        ).run_id
        _ = instance.create_run_for_pipeline(
            repo.get_pipeline("foo_pipeline"), status=PipelineRunStatus.SUCCESS
        ).run_id
        with define_out_of_process_context(__file__, "get_repo", instance) as context:
            result = execute_dagster_graphql(context, FILTERED_RUNS_QUERY)
            assert result.data
            run_ids = [run["runId"] for run in result.data["pipelineRunsOrError"]["results"]]
            assert len(run_ids) == 1
            assert run_ids[0] == run_id_2


def test_repositories_query():
    with instance_for_test() as instance:
        with define_out_of_process_context(__file__, "get_repo", instance) as context:
            result = execute_dagster_graphql(context, REPOSITORIES_QUERY)
            assert not result.errors
            assert result.data
            repositories = result.data["repositoriesOrError"]["nodes"]
            assert len(repositories) == 1
            assert repositories[0]["name"] == "my_repo"


def test_pipelines_query():
    with instance_for_test() as instance:
        with define_out_of_process_context(__file__, "get_repo", instance) as context:
            result = execute_dagster_graphql(
                context,
                PIPELINES_QUERY,
                variables={
                    "repositoryLocationName": "test_location",
                    "repositoryName": "my_repo",
                },
            )
            assert not result.errors
            assert result.data
            pipelines = result.data["repositoryOrError"]["pipelines"]
            assert len(pipelines) == 2


def test_launch_mutation():
    with instance_for_test() as instance:
        with define_out_of_process_context(__file__, "get_repo", instance) as context:
            result = execute_dagster_graphql(
                context,
                LAUNCH_PIPELINE,
                variables={
                    "repositoryLocationName": "test_location",
                    "repositoryName": "my_repo",
                    "pipelineName": "foo_pipeline",
                    "runConfigData": {},
                    "mode": "default",
                },
            )
            assert not result.errors
            assert result.data
            run = result.data["launchPipelineExecution"]["run"]
            assert run
            assert run["runId"]


def test_launch_mutation_error():
    with instance_for_test() as instance:
        with define_out_of_process_context(__file__, "get_repo", instance) as context:
            result = execute_dagster_graphql(
                context,
                LAUNCH_PIPELINE,
                variables={
                    "repositoryLocationName": "test_location",
                    "repositoryName": "my_repo",
                    "pipelineName": "foo_pipeline",
                    "runConfigData": {"invalid": "config"},
                    "mode": "default",
                },
            )
            assert not result.errors
            assert result.data
            errors = result.data["launchPipelineExecution"]["errors"]
            assert len(errors) == 1
            message = errors[0]["message"]
            assert message


def test_launch_preset_mutation():
    with instance_for_test() as instance:
        with define_out_of_process_context(__file__, "get_repo", instance) as context:
            result = execute_dagster_graphql(
                context,
                LAUNCH_PIPELINE_PRESET,
                variables={
                    "repositoryLocationName": "test_location",
                    "repositoryName": "my_repo",
                    "pipelineName": "foo_pipeline",
                    "presetName": "my_preset",
                },
            )
            assert not result.errors
            assert result.data
            run = result.data["launchPipelineExecution"]["run"]
            assert run
            assert run["runId"]



import time

import pytest
from dagster_graphql.test.utils import execute_dagster_graphql
from dagster_graphql_tests.graphql.graphql_context_test_suite import (
    ExecutingGraphQLContextTestMatrix,
)

from dagster.core.scheduler import DagsterDaemonScheduler
from dagster.daemon.daemon import SensorDaemon
from dagster.daemon.types import DaemonHeartbeat
from dagster.utils.error import SerializableErrorInfo

INDIVIDUAL_DAEMON_QUERY = """
query InstanceDetailSummaryQuery {
    instance {
        daemonHealth {
            id
            sensor: daemonStatus(daemonType: "SENSOR") {
                daemonType
                required
                healthy
                lastHeartbeatTime
            }
            run_coordinator: daemonStatus(daemonType: "QUEUED_RUN_COORDINATOR") {
                daemonType
                required
                healthy
                lastHeartbeatTime
            }
            scheduler: daemonStatus(daemonType: "SCHEDULER") {
                daemonType
                required
                healthy
                lastHeartbeatTime
            }
        }
    }
}
"""

ALL_DAEMON_QUERY = """
query InstanceDetailSummaryQuery {
    instance {
        daemonHealth {
            id
            allDaemonStatuses {
                daemonType
                required
                healthy
                lastHeartbeatTime
            }
        }
    }
}
"""

DAEMON_HEALTH_QUERY = """
query InstanceDetailSummaryQuery {
    instance {
        daemonHealth {
            id
            sensor: daemonStatus(daemonType: "SENSOR"){
                lastHeartbeatErrors {
                    message
                }
                healthy
            }
        }
    }
}
"""


class TestDaemonHealth(ExecutingGraphQLContextTestMatrix):
    def test_get_individual_daemons(self, graphql_context):
        if graphql_context.instance.is_ephemeral:
            pytest.skip("The daemon isn't compatible with an in-memory instance")
        graphql_context.instance.add_daemon_heartbeat(
            DaemonHeartbeat(
                timestamp=100.0, daemon_type=SensorDaemon.daemon_type(), daemon_id=None, errors=None
            )
        )
        results = execute_dagster_graphql(graphql_context, INDIVIDUAL_DAEMON_QUERY)

        scheduler_required = isinstance(graphql_context.instance.scheduler, DagsterDaemonScheduler)
        assert results.data == {
            "instance": {
                "daemonHealth": {
                    "id": "daemonHealth",
                    "sensor": {
                        "daemonType": "SENSOR",
                        "required": True,
                        "healthy": False,
                        "lastHeartbeatTime": 100.0,
                    },
                    "run_coordinator": {
                        "daemonType": "QUEUED_RUN_COORDINATOR",
                        "required": False,
                        "healthy": None,
                        "lastHeartbeatTime": None,
                    },
                    "scheduler": {
                        "daemonType": "SCHEDULER",
                        "required": scheduler_required,
                        "healthy": False if scheduler_required else None,
                        "lastHeartbeatTime": None,
                    },
                }
            }
        }

    def test_get_all_daemons(self, graphql_context):
        if graphql_context.instance.is_ephemeral:
            pytest.skip("The daemon isn't compatible with an in-memory instance")
        results = execute_dagster_graphql(graphql_context, ALL_DAEMON_QUERY)
        scheduler_required = isinstance(graphql_context.instance.scheduler, DagsterDaemonScheduler)

        assert results.data == {
            "instance": {
                "daemonHealth": {
                    "id": "daemonHealth",
                    "allDaemonStatuses": [
                        {
                            "daemonType": "SENSOR",
                            "required": True,
                            "healthy": False,
                            "lastHeartbeatTime": None,
                        },
                        {
                            "daemonType": "BACKFILL",
                            "required": True,
                            "healthy": False,
                            "lastHeartbeatTime": None,
                        },
                    ]
                    + (
                        [
                            {
                                "daemonType": "SCHEDULER",
                                "required": True,
                                "healthy": False if scheduler_required else None,
                                "lastHeartbeatTime": None,
                            }
                        ]
                        if scheduler_required
                        else []
                    ),
                }
            }
        }

    def test_get_daemon_error(self, graphql_context):
        if graphql_context.instance.is_ephemeral:
            pytest.skip("The daemon isn't compatible with an in-memory instance")
        graphql_context.instance.add_daemon_heartbeat(
            DaemonHeartbeat(
                timestamp=time.time(),
                daemon_type=SensorDaemon.daemon_type(),
                daemon_id=None,
                errors=[
                    SerializableErrorInfo(message="foobar", stack=[], cls_name=None, cause=None)
                ],
            )
        )
        results = execute_dagster_graphql(graphql_context, DAEMON_HEALTH_QUERY)
        assert results.data["instance"]["daemonHealth"]["sensor"] == {
            "lastHeartbeatErrors": [{"message": "foobar"}],
            "healthy": True,
        }



import time
from typing import Any

from dagster.core.host_representation.grpc_server_state_subscriber import (
    LocationStateChangeEvent,
    LocationStateChangeEventType,
    LocationStateSubscriber,
)

from .graphql_context_test_suite import GraphQLContextVariant, make_graphql_context_test_suite

BaseTestSuite: Any = make_graphql_context_test_suite(
    context_variants=[GraphQLContextVariant.non_launchable_sqlite_instance_deployed_grpc_env()]
)


class TestSubscribeToGrpcServerEvents(BaseTestSuite):
    def test_grpc_server_handle_message_subscription(self, graphql_context):
        events = []
        test_subscriber = LocationStateSubscriber(events.append)
        location = next(
            iter(
                graphql_context.process_context.create_request_context().repository_locations  # pylint: disable=protected-access
            )
        )
        graphql_context.process_context.add_state_subscriber(  # pylint: disable=protected-access
            test_subscriber
        )
        location.client.shutdown_server()

        # Wait for event
        start_time = time.time()
        timeout = 60
        while not len(events) > 0:
            if time.time() - start_time > timeout:
                raise Exception("Timed out waiting for LocationStateChangeEvent")
            time.sleep(1)

        assert len(events) == 1
        assert isinstance(events[0], LocationStateChangeEvent)
        assert events[0].event_type == LocationStateChangeEventType.LOCATION_ERROR
        assert events[0].location_name == location.name



import graphene

from dagster import check
from dagster.config.config_type import ConfigTypeKind
from dagster.config.snap import get_recursive_type_keys
from dagster.core.snap import ConfigFieldSnap, ConfigSchemaSnapshot, ConfigTypeSnap

from .util import non_null_list


def to_config_type(config_schema_snapshot, config_type_key):
    check.inst_param(config_schema_snapshot, "config_schema_snapshot", ConfigSchemaSnapshot)
    check.str_param(config_type_key, "config_type_key")

    config_type_snap = config_schema_snapshot.get_config_snap(config_type_key)
    kind = config_type_snap.kind

    if kind == ConfigTypeKind.ENUM:
        return GrapheneEnumConfigType(config_schema_snapshot, config_type_snap)
    elif ConfigTypeKind.has_fields(kind):
        return GrapheneCompositeConfigType(config_schema_snapshot, config_type_snap)
    elif kind == ConfigTypeKind.ARRAY:
        return GrapheneArrayConfigType(config_schema_snapshot, config_type_snap)
    elif kind == ConfigTypeKind.MAP:
        return GrapheneMapConfigType(config_schema_snapshot, config_type_snap)
    elif kind == ConfigTypeKind.NONEABLE:
        return GrapheneNullableConfigType(config_schema_snapshot, config_type_snap)
    elif kind == ConfigTypeKind.ANY or kind == ConfigTypeKind.SCALAR:
        return GrapheneRegularConfigType(config_schema_snapshot, config_type_snap)
    elif kind == ConfigTypeKind.SCALAR_UNION:
        return GrapheneScalarUnionConfigType(config_schema_snapshot, config_type_snap)
    else:
        check.failed("Should never reach")


def _ctor_kwargs_for_snap(config_type_snap):
    return dict(
        key=config_type_snap.key,
        description=config_type_snap.description,
        is_selector=config_type_snap.kind == ConfigTypeKind.SELECTOR,
        type_param_keys=config_type_snap.type_param_keys or [],
    )


class GrapheneConfigType(graphene.Interface):
    key = graphene.NonNull(graphene.String)
    description = graphene.String()

    recursive_config_types = graphene.Field(
        non_null_list(lambda: GrapheneConfigType),
        description="""
This is an odd and problematic field. It recursively goes down to
get all the types contained within a type. The case where it is horrible
are dictionaries and it recurses all the way down to the leaves. This means
that in a case where one is fetching all the types and then all the inner
types keys for those types, we are returning O(N^2) type keys, which
can cause awful performance for large schemas. When you have access
to *all* the types, you should instead only use the type_param_keys
field for closed generic types and manually navigate down the to
field types client-side.

Where it is useful is when you are fetching types independently and
want to be able to render them, but without fetching the entire schema.

We use this capability when rendering the sidebar.
    """,
    )
    type_param_keys = graphene.Field(
        non_null_list(graphene.String),
        description="""
This returns the keys for type parameters of any closed generic type,
(e.g. List, Optional). This should be used for reconstructing and
navigating the full schema client-side and not innerTypes.
    """,
    )
    is_selector = graphene.NonNull(graphene.Boolean)

    class Meta:
        name = "ConfigType"


class ConfigTypeMixin:
    def __init__(self, config_schema_snapshot, config_type_snap):
        self._config_type_snap = check.inst_param(
            config_type_snap, "config_type_snap", ConfigTypeSnap
        )
        self._config_schema_snapshot = check.inst_param(
            config_schema_snapshot, "config_schema_snapshot", ConfigSchemaSnapshot
        )
        super().__init__(**_ctor_kwargs_for_snap(config_type_snap))

    def resolve_recursive_config_types(self, _graphene_info):
        return list(
            map(
                lambda key: to_config_type(self._config_schema_snapshot, key),
                get_recursive_type_keys(self._config_type_snap, self._config_schema_snapshot),
            )
        )


class GrapheneRegularConfigType(ConfigTypeMixin, graphene.ObjectType):
    class Meta:
        interfaces = (GrapheneConfigType,)
        description = "Regular is an odd name in this context. It really means Scalar or Any."
        name = "RegularConfigType"

    given_name = graphene.NonNull(graphene.String)

    def resolve_given_name(self, _):
        return self._config_type_snap.given_name


class GrapheneMapConfigType(ConfigTypeMixin, graphene.ObjectType):

    key_type = graphene.Field(graphene.NonNull(GrapheneConfigType))
    value_type = graphene.Field(graphene.NonNull(GrapheneConfigType))
    key_label_name = graphene.Field(graphene.String)

    class Meta:
        interfaces = (GrapheneConfigType,)
        name = "MapConfigType"

    def resolve_key_type(self, _graphene_info):
        return to_config_type(
            self._config_schema_snapshot,
            self._config_type_snap.key_type_key,
        )

    def resolve_value_type(self, _graphene_info):
        return to_config_type(
            self._config_schema_snapshot,
            self._config_type_snap.inner_type_key,
        )

    def resolve_key_label_name(self, _graphene_info):
        return self._config_type_snap.given_name


class GrapheneWrappingConfigType(graphene.Interface):
    of_type = graphene.Field(graphene.NonNull(GrapheneConfigType))

    class Meta:
        name = "WrappingConfigType"


class GrapheneArrayConfigType(ConfigTypeMixin, graphene.ObjectType):
    class Meta:
        interfaces = (GrapheneConfigType, GrapheneWrappingConfigType)
        name = "ArrayConfigType"

    def resolve_of_type(self, _graphene_info):
        return to_config_type(
            self._config_schema_snapshot,
            self._config_type_snap.inner_type_key,
        )


class GrapheneScalarUnionConfigType(ConfigTypeMixin, graphene.ObjectType):
    scalar_type = graphene.NonNull(GrapheneConfigType)
    non_scalar_type = graphene.NonNull(GrapheneConfigType)
    scalar_type_key = graphene.NonNull(graphene.String)
    non_scalar_type_key = graphene.NonNull(graphene.String)

    class Meta:
        interfaces = (GrapheneConfigType,)
        name = "ScalarUnionConfigType"

    def get_scalar_type_key(self):
        return self._config_type_snap.scalar_type_key

    def get_non_scalar_type_key(self):
        return self._config_type_snap.non_scalar_type_key

    def resolve_scalar_type_key(self, _):
        return self.get_scalar_type_key()

    def resolve_non_scalar_type_key(self, _):
        return self.get_non_scalar_type_key()

    def resolve_scalar_type(self, _):
        return to_config_type(self._config_schema_snapshot, self.get_scalar_type_key())

    def resolve_non_scalar_type(self, _):
        return to_config_type(self._config_schema_snapshot, self.get_non_scalar_type_key())


class GrapheneNullableConfigType(ConfigTypeMixin, graphene.ObjectType):
    class Meta:
        interfaces = (GrapheneConfigType, GrapheneWrappingConfigType)
        name = "NullableConfigType"

    def resolve_of_type(self, _graphene_info):
        return to_config_type(self._config_schema_snapshot, self._config_type_snap.inner_type_key)


class GrapheneEnumConfigValue(graphene.ObjectType):
    value = graphene.NonNull(graphene.String)
    description = graphene.String()

    class Meta:
        name = "EnumConfigValue"


class GrapheneEnumConfigType(ConfigTypeMixin, graphene.ObjectType):
    class Meta:
        interfaces = (GrapheneConfigType,)
        name = "EnumConfigType"

    values = non_null_list(GrapheneEnumConfigValue)
    given_name = graphene.NonNull(graphene.String)

    def resolve_values(self, _graphene_info):
        return [
            GrapheneEnumConfigValue(value=ev.value, description=ev.description)
            for ev in self._config_type_snap.enum_values
        ]

    def resolve_given_name(self, _):
        return self._config_type_snap.given_name


class GrapheneConfigTypeField(graphene.ObjectType):
    name = graphene.NonNull(graphene.String)
    description = graphene.String()
    config_type = graphene.NonNull(GrapheneConfigType)
    config_type_key = graphene.NonNull(graphene.String)
    is_required = graphene.NonNull(graphene.Boolean)

    class Meta:
        name = "ConfigTypeField"

    def resolve_config_type_key(self, _):
        return self._field_snap.type_key

    def __init__(self, config_schema_snapshot, field_snap):
        self._config_schema_snapshot = check.inst_param(
            config_schema_snapshot, "config_schema_snapshot", ConfigSchemaSnapshot
        )
        self._field_snap = check.inst_param(field_snap, "field_snap", ConfigFieldSnap)
        super().__init__(
            name=field_snap.name,
            description=field_snap.description,
            is_required=field_snap.is_required,
        )

    def resolve_config_type(self, _graphene_info):
        return to_config_type(self._config_schema_snapshot, self._field_snap.type_key)


class GrapheneCompositeConfigType(ConfigTypeMixin, graphene.ObjectType):
    fields = non_null_list(GrapheneConfigTypeField)

    class Meta:
        interfaces = (GrapheneConfigType,)
        name = "CompositeConfigType"

    def resolve_fields(self, _graphene_info):
        return sorted(
            [
                GrapheneConfigTypeField(
                    config_schema_snapshot=self._config_schema_snapshot,
                    field_snap=field_snap,
                )
                for field_snap in self._config_type_snap.fields
            ],
            key=lambda field: field.name,
        )


types = [
    GrapheneArrayConfigType,
    GrapheneCompositeConfigType,
    GrapheneConfigType,
    GrapheneConfigTypeField,
    GrapheneEnumConfigType,
    GrapheneEnumConfigValue,
    GrapheneNullableConfigType,
    GrapheneRegularConfigType,
    GrapheneScalarUnionConfigType,
    GrapheneWrappingConfigType,
    GrapheneMapConfigType,
]



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot

snapshots = Snapshot()

snapshots['test_query_inputs_outputs 1'] = {
    'repositoryOrError': {
        'usedSolid': {
            '__typename': 'UsedSolid',
            'definition': {
                'inputDefinitions': [
                    {
                        'metadataEntries': [
                            {
                                'label': 'a'
                            }
                        ]
                    }
                ],
                'outputDefinitions': [
                    {
                        'metadataEntries': [
                            {
                                'label': 'c'
                            }
                        ]
                    }
                ]
            }
        }
    }
}



from .client import DagsterGraphQLClient
from .utils import (
    DagsterGraphQLClientError,
    InvalidOutputErrorInfo,
    ReloadRepositoryLocationInfo,
    ReloadRepositoryLocationStatus,
    ShutdownRepositoryLocationInfo,
    ShutdownRepositoryLocationStatus,
)

__all__ = [
    "DagsterGraphQLClient",
    "DagsterGraphQLClientError",
    "ReloadRepositoryLocationInfo",
    "ReloadRepositoryLocationStatus",
    "ShutdownRepositoryLocationInfo",
    "ShutdownRepositoryLocationStatus",
]



from typing import Any

import pytest
from dagster_graphql import DagsterGraphQLClientError, ReloadRepositoryLocationStatus

from ..graphql.graphql_context_test_suite import (
    GraphQLContextVariant,
    make_graphql_context_test_suite,
)
from .conftest import MockClient, python_client_test_suite


@python_client_test_suite
def test_success(mock_client: MockClient):
    response = {
        "reloadRepositoryLocation": {
            "__typename": "WorkspaceLocationEntry",
            "locationOrLoadError": {"__typename": "RepositoryLocation"},
        }
    }
    mock_client.mock_gql_client.execute.return_value = response

    assert (
        mock_client.python_client.reload_repository_location("foo").status
        == ReloadRepositoryLocationStatus.SUCCESS
    )


@python_client_test_suite
def test_failure_with_repo_location_load_failure(mock_client: MockClient):
    error_type, error_msg = "PythonError", "some reason"
    response = {
        "reloadRepositoryLocation": {
            "__typename": "WorkspaceLocationEntry",
            "locationOrLoadError": {
                "__typename": error_type,
                "message": error_msg,
            },
        }
    }
    mock_client.mock_gql_client.execute.return_value = response

    result = mock_client.python_client.reload_repository_location("foo")
    assert result.status == ReloadRepositoryLocationStatus.FAILURE
    assert result.failure_type == error_type
    assert result.message == error_msg


@python_client_test_suite
def test_failure_with_reload_not_supported(mock_client: MockClient):
    error_type, error_msg = "ReloadNotSupported", "some reason"
    response = {
        "reloadRepositoryLocation": {
            "__typename": error_type,
            "message": error_msg,
        }
    }
    mock_client.mock_gql_client.execute.return_value = response

    result = mock_client.python_client.reload_repository_location("foo")
    assert result.status == ReloadRepositoryLocationStatus.FAILURE
    assert result.failure_type == error_type
    assert result.message == error_msg


@python_client_test_suite
def test_failure_with_repo_location_not_found(mock_client: MockClient):
    error_type, error_msg = "RepositoryLocationNotFound", "some reason"
    response = {
        "reloadRepositoryLocation": {
            "__typename": error_type,
            "message": error_msg,
        }
    }
    mock_client.mock_gql_client.execute.return_value = response

    result = mock_client.python_client.reload_repository_location("foo")
    assert result.status == ReloadRepositoryLocationStatus.FAILURE
    assert result.failure_type == error_type
    assert result.message == error_msg


@python_client_test_suite
def test_failure_with_query_error(mock_client: MockClient):
    mock_client.mock_gql_client.execute.side_effect = Exception("foo")

    with pytest.raises(DagsterGraphQLClientError) as _:
        mock_client.python_client.reload_repository_location("foo")


BaseTestSuite: Any = make_graphql_context_test_suite(
    context_variants=[GraphQLContextVariant.non_launchable_in_memory_instance_managed_grpc_env()]
)


class TestReloadRepositoryLocationWithClient(BaseTestSuite):
    def test_reload_location_real(self, graphql_client):
        assert (
            graphql_client.reload_repository_location("test").status
            == ReloadRepositoryLocationStatus.SUCCESS
        )
